search.o search.d : search.c genesis/sbcl.h interr.h os.h runtime.h lispobj.h \
 align.h target-os.h target-arch-os.h target-arch.h search.h thread.h \
 globals.h genesis/symbol.h lispobj.h symbol-tls.inc align.h \
 genesis/vector.h genesis/static-symbols.h genesis/thread.h \
 gencgc-alloc-region.h genesis/vector.h interrupt.h validate.h \
 gc-typedefs.h gc.h gc-assert.h immobile-space.h gencgc-impl.h \
 gencgc-alloc-region.h genesis/cardmarks.h genesis/cons.h \
 genesis/weak-pointer.h hopscotch.h code.h genesis/closure.h \
 forwarding-ptr.inc genesis/primitive-objects.h genesis/number-types.h \
 genesis/closure.h genesis/cons.h genesis/instance.h instance.inc \
 genesis/sap.h genesis/symbol.h genesis/thread.h genesis/value-cell.h \
 genesis/vector.h genesis/weak-pointer.h genesis/instance.h \
 genesis/hash-table.h genesis/package.h genesis/split-ordered-list.h \
 genesis/brothertree.h
