interr.o interr.d : interr.c genesis/sbcl.h arch.h os.h runtime.h lispobj.h align.h \
 target-os.h target-arch-os.h target-arch.h thread.h globals.h \
 genesis/symbol.h lispobj.h symbol-tls.inc align.h genesis/vector.h \
 genesis/static-symbols.h genesis/thread.h gencgc-alloc-region.h \
 genesis/vector.h interrupt.h validate.h gc-typedefs.h interr.h print.h \
 lispregs.h target-lispregs.h genesis/regnames.h code.h genesis/closure.h \
 gc-assert.h breakpoint.h var-io.h sc-offset.h gc.h immobile-space.h \
 gencgc-impl.h gencgc-alloc-region.h genesis/cardmarks.h genesis/cons.h \
 genesis/weak-pointer.h hopscotch.h forwarding-ptr.inc genesis/errnames.h
