immobile-space.o immobile-space.d : immobile-space.c code.h genesis/closure.h lispobj.h \
 genesis/sbcl.h gc-assert.h interr.h os.h genesis/sbcl.h runtime.h \
 lispobj.h align.h target-os.h target-arch-os.h target-arch.h core.h gc.h \
 gc-typedefs.h globals.h immobile-space.h gencgc-impl.h \
 gencgc-alloc-region.h genesis/cardmarks.h genesis/cons.h \
 genesis/weak-pointer.h hopscotch.h forwarding-ptr.inc \
 genesis/brothertree.h genesis/gc-tables.h genesis/hash-table.h \
 genesis/instance.h instance.inc align.h genesis/package.h \
 genesis/static-symbols.h genesis/symbol.h symbol-tls.inc \
 genesis/vector.h genesis/vector.h pseudo-atomic.h genesis/thread.h \
 gencgc-alloc-region.h search.h validate.h thread.h interrupt.h var-io.h \
 ../../tlsf-bsd/tlsf/tlsf.h
