linux-os.o linux-os.d : linux-os.c genesis/sbcl.h os.h runtime.h lispobj.h align.h \
 target-os.h target-arch-os.h target-arch.h arch.h thread.h globals.h \
 genesis/symbol.h lispobj.h symbol-tls.inc align.h genesis/vector.h \
 genesis/static-symbols.h genesis/thread.h gencgc-alloc-region.h \
 genesis/vector.h interrupt.h validate.h gc-typedefs.h interr.h \
 lispregs.h target-lispregs.h genesis/regnames.h genesis/events.h gc.h \
 gc-assert.h immobile-space.h gencgc-impl.h gencgc-alloc-region.h \
 genesis/cardmarks.h genesis/cons.h genesis/weak-pointer.h hopscotch.h \
 code.h genesis/closure.h forwarding-ptr.inc
